from typing import Literal

import torch
from torch import Tensor, nn


class ConceptEraser(nn.Module):
    """Minimally edit features to make specified concepts linearly undetectable."""

    mean_x: Tensor
    """Running mean of X."""

    mean_z: Tensor
    """Running mean of Z."""

    sigma_xz_M2: Tensor
    """Unnormalized cross-covariance matrix X^T Z."""

    x_M2: Tensor | None
    """Unnormalized covariance matrix X^T X."""

    n_x: Tensor
    """Number of X samples seen so far."""

    n_z: Tensor
    """Number of Z samples seen so far."""

    _P: Tensor | None

    @classmethod
    def fit(cls, x: Tensor, y: Tensor, **kwargs) -> "ConceptEraser":
        """Convenience method to fit a ConceptEraser on data and return it."""
        n, d = x.shape
        _, k = y.reshape(n, -1).shape

        return cls(d, k, device=x.device, dtype=x.dtype, **kwargs).update(x, y)

    def __init__(
        self,
        x_dim: int,
        z_dim: int,
        proj_type: Literal["leace", "orth", "relaxed"] = "leace",
        *,
        affine: bool = True,
        device: str | torch.device | None = None,
        dtype: torch.dtype | None = None,
        svd_tol: float = 0.01,
    ):
        """Initialize a ConceptEraser.

        Args:
            x_dim: Dimensionality of the input.
            z_dim: Dimensionality of the labels.
            proj_type: Type of projection matrix to use.
            affine: Whether to use a bias term to ensure the unconditional mean of the
                features remains the same after erasure.
            device: Device to put the statistics on.
            dtype: Data type to use for the statistics.
            svd_tol: Singular values under this threshold are truncated, both during
                the phase where we do SVD on the cross-covariance matrix, and at the
                phase where we compute the pseudoinverse of the projected covariance
                matrix. Higher values are more numerically stable and result in less
                damage to the representation, but may leave trace correlations intact.
        """
        super().__init__()

        self.z_dim = z_dim
        self.x_dim = x_dim

        self.affine = affine
        self.proj_rank = z_dim
        self.proj_type = proj_type
        self.z_dim = z_dim

        assert svd_tol > 0.0, "`svd_tol` must be positive for numerical stability."
        self.svd_tol = svd_tol

        self.register_buffer("mean_x", torch.zeros(x_dim, device=device, dtype=dtype))
        self.register_buffer("mean_z", self.mean_x.new_zeros(z_dim))
        self.register_buffer(
            "sigma_xz_M2",
            self.mean_x.new_zeros(x_dim, z_dim),
        )
        self.register_buffer("n_x", torch.tensor(0, device=device, dtype=dtype))
        self.register_buffer("n_z", torch.tensor(0, device=device, dtype=dtype))
        self.register_buffer("_P", None)

        if self.proj_type == "leace":
            M2 = self.mean_x.new_zeros(x_dim, x_dim)
        elif self.proj_type in ("orth", "relaxed"):
            M2 = None
        else:
            raise ValueError(f"Unknown projection type {self.proj_type}")

        self.register_buffer("x_M2", M2)

    def forward(self, x: Tensor) -> Tensor:
        """Minimally edit `x` to remove correlations with the target concepts.

        Args:
            x: Representations of shape (..., x_dim).

        Returns:
            The edited representations of shape (..., x_dim).
        """
        d, _ = self.sigma_xz_M2.shape
        assert self.n_x > 0, "Call update() before forward()"
        assert x.shape[-1] == d

        if self.affine:
            x_ = (x - self.mean_x) @ self.P.T + self.mean_x
            return x_.type_as(x)
        else:
            return (x.type_as(self.P) @ self.P.T).type_as(x)

    @torch.no_grad()
    def update(self, x: Tensor, z: Tensor | None = None) -> "ConceptEraser":
        """Update the running statistics with a new batch of data.

        It's possible to call this method without `z` if you only want to update the
        statistics of X. This is useful if you don't have labels but want to adjust the
        mean and covariance of X to match a new dataset.
        """
        d, c = self.sigma_xz_M2.shape
        x = x.reshape(-1, d).type_as(self.mean_x)

        if not x.isfinite().all():
            raise RuntimeError("Non-finite values in input")

        n, d2 = x.shape
        assert d == d2, f"Unexpected number of features {d2}"

        # We always have an X, we might not have a Z
        self.n_x += n

        # Welford's online algorithm
        delta_x = x - self.mean_x
        self.mean_x += delta_x.sum(dim=0) / self.n_x
        delta_x2 = x - self.mean_x

        # Update the covariance matrix of X if needed (for LEACE)
        if self.proj_type == "leace":
            assert self.x_M2 is not None
            self.x_M2.addmm_(delta_x.mT, delta_x2)

        # Invalidate the cached projection matrix
        self._P = None

        # We do have labels, so we can update the Z statistics
        if z is not None:
            # y might start out 1D, but we want to treat it as 2D
            z = z.reshape(n, -1).type_as(x)
            assert z.shape[-1] == c, f"Unexpected number of classes {z.shape[-1]}"

            self.n_z += n

            delta_z = z - self.mean_z
            self.mean_z += delta_z.sum(dim=0) / self.n_x
            delta_z2 = z - self.mean_z

            # Update the cross-covariance matrix
            self.sigma_xz_M2.addmm_(delta_x.mT, delta_z2)

        return self

    @property
    def P(self) -> Tensor:
        """Projection matrix for removing the subspace."""
        if self._P is not None:
            return self._P

        eye = torch.eye(self.x_dim, device=self.mean_x.device, dtype=self.mean_x.dtype)
        u, s, _ = torch.linalg.svd(self.sigma_xz, full_matrices=False)

        if self.proj_type == "relaxed":
            # Treat `svd_tol` as a constraint on the spectral norm of sigma_xz after
            # erasure. In this case Q is not actually a projection matrix, it's a
            # PSD non-expansive map (spectral norm <= 1).
            Q = eye - u * torch.clamp(1 - self.svd_tol / s, 0, 1) @ u.T
        else:
            # Throw away singular values that are too small
            mask = s > self.svd_tol
            if not mask.any():
                return eye

            u = u[:, mask]
            Q = eye - u @ u.T if mask.any() else eye

            # Save this for debugging
            self.proj_rank = mask.sum().item()

        if self.proj_type != "leace":
            self._P = Q
            return self._P

        self._P = self.proj_for_subspace(u)
        return self._P

    def proj_for_subspace(self, u: Tensor) -> Tensor:
        """Projection matrix for removing the subspace."""
        eye = torch.eye(self.x_dim, device=self.mean_x.device, dtype=self.mean_x.dtype)
        Q = eye - u @ u.T

        # LEACE and orthogonal projection matrix computation
        # Adjust Q to account for the covariance of X
        sigma = self.cov_x
        A = Q @ sigma @ Q
        try:
            L, V = torch.linalg.eigh(A)
        except torch.linalg.LinAlgError as e:
            # Better error messages in the common case where A is non-finite
            if not A.isfinite().all():
                raise RuntimeError("Non-finite values in covariance matrix") from e
            else:
                raise e

        # Manual pseudoinverse
        L = L.reciprocal().where(L > self.svd_tol, 0.0)
        P = sigma @ V @ torch.diag_embed(L) @ V.mT

        # Prevent the covariance trace from increasing
        old_trace = torch.trace(sigma)
        new_trace = torch.trace(P @ sigma @ P.mT)

        # If applying the projection matrix increases the variance, this might
        # cause instability, especially when erasure is applied multiple times.
        # We regularize toward the orthogonal projection matrix to avoid this.
        if new_trace > old_trace:
            # Set up the variables for the quadratic equation
            x = new_trace
            y = 2 * torch.trace(P @ sigma @ Q.mT)
            z = torch.trace(Q @ sigma @ Q.mT)
            w = old_trace

            # Solve for the mixture of P and Q that makes the trace equal to the
            # trace of the original covariance matrix
            discr = torch.sqrt(4 * w * x - 4 * w * y + 4 * w * z - 4 * x * z + y**2)
            alpha1 = (-y / 2 + z - discr / 2) / (x - y + z)
            alpha2 = (-y / 2 + z + discr / 2) / (x - y + z)

            # Choose the positive root
            alpha = torch.where(alpha1 > 0, alpha1, alpha2).clamp(0, 1)
            P = alpha * P + (1 - alpha) * Q

        return P

    @property
    def cov_x(self) -> Tensor:
        """The covariance matrix of X."""
        assert self.n_x > 1, "Call update() before accessing cov_x"
        assert (
            self.x_M2 is not None
        ), "Covariance statistics are not being tracked for X"

        cov = self.x_M2 / (self.n_x - 1)

        # Accumulated numerical error may cause this to be slightly non-symmetric
        return (cov + cov.mT) / 2

    @property
    def sigma_xz(self) -> Tensor:
        """The cross-covariance matrix."""
        assert self.n_z > 1, "Call update() with labels before accessing sigma_xz"
        return self.sigma_xz_M2 / (self.n_z - 1)

    def finalize(self) -> "ConceptEraser":
        """Compute the projection matrix and drop covariance matrices."""
        self.P
        self.x_M2 = None
        return self



from types import MethodType
from typing import Literal

import torch
import torch.nn.functional as F
from datasets import Dataset
from tqdm.auto import tqdm
from transformers import (
    GPTNeoXForCausalLM,
    GPTNeoXLayer,
    GPTNeoXModel,
)
from transformers.models.gpt_neox.modeling_gpt_neox import GPTNeoXAttention

from concept_erasure import ConceptEraser, ConceptScrubber
from concept_erasure.utils import assert_type


def patch_attention_neox_(model: torch.nn.Module):
    """Patch the attention layers of a GPTNeoX model to use fast attention kernels."""

    def fast_attn(self, q, k, v, attention_mask=None, head_mask=None):
        assert attention_mask is None, "attention_mask not implemented for fast_attn"
        assert head_mask is None, "head_mask not implemented for fast_attn"

        # Queries and keys are always float32 because of the rotary embeddings
        q, k = q.type_as(v), k.type_as(v)
        return F.scaled_dot_product_attention(q, k, v, is_causal=True), None

    for mod in model.modules():
        if isinstance(mod, GPTNeoXAttention):
            mod._attn = MethodType(fast_attn, mod)


@torch.no_grad()
def scrub_neox(
    model: GPTNeoXForCausalLM,
    train: Dataset,
    z_column: str | None,
    batch_size: int = 32,
    proj_type: Literal["leace", "loracs", "orth"] = "leace",
    affine: bool = True,
) -> tuple[ConceptScrubber | None, float]:
    base = assert_type(GPTNeoXModel, model.base_model)
    d = assert_type(int, base.config.hidden_size)

    # Put model in eval mode
    model.eval()
    model.requires_grad_(False)

    if z_column is None:
        k = -1
        scrubber = None
    else:
        k = assert_type(int, train.features[z_column].feature.num_classes)
        scrubber = ConceptScrubber()

    losses = []
    xs = []
    zs = []
    N = len(train) // batch_size
    train = train.with_format("torch", device=model.device)

    # Embed all the inputs
    embed_fn = base.get_input_embeddings()
    for i, batch in enumerate(tqdm(train.iter(batch_size), desc="Embedding", total=N)):
        assert isinstance(batch, dict)

        tokens = assert_type(torch.Tensor, batch["input_ids"])
        x = embed_fn(tokens)
        xs.append(x.to("cpu", non_blocking=True))

        # We don't actually need to move these to the CPU since they're small
        if z_column is not None:
            zs.append(F.one_hot(batch[z_column], num_classes=k))

    # Enumerate the layers
    for j, layer in enumerate(tqdm(base.layers, unit="layer")):
        assert isinstance(layer, GPTNeoXLayer)
        assert layer.use_parallel_residual, "Only parallel residual is supported"

        attn_eraser, mlp_eraser = None, None
        if scrubber is not None:
            attn_eraser = ConceptEraser(
                d, k, affine=affine, device=model.device, proj_type=proj_type
            )
            scrubber.erasers[f"layers-{j}-input_layernorm"] = attn_eraser

            mlp_eraser = ConceptEraser(
                d, k, affine=affine, device=model.device, proj_type=proj_type
            )
            scrubber.erasers[f"layers-{j}-post_attention_layernorm"] = mlp_eraser

            # Fit the next eraser on the previous hidden states
            for i, (x, z) in tqdm(enumerate(zip(xs, zs)), desc="Fitting", total=N):
                x = x.to(model.device)

                # Discard post-LN output and recompute during application to save RAM
                attn_norm_out = layer.input_layernorm(x)
                attn_eraser.update(attn_norm_out, z)

                mlp_norm_out = layer.post_attention_layernorm(attn_norm_out)
                mlp_eraser.update(mlp_norm_out, z)

            # Save VRAM by discarding the covariance matrices
            attn_eraser.finalize()
            mlp_eraser.finalize()

        # Run attention & MLP with the erasers we just fit
        for i, x in tqdm(enumerate(xs), desc="Applying", total=N):
            x = x.to(model.device)
            h = layer.input_layernorm(x)  # Recomputing from above

            if attn_eraser is not None:
                h = attn_eraser(h)

            pos_ids = torch.arange(0, x.shape[-2], device=x.device, dtype=torch.long)
            pos_ids = pos_ids.unsqueeze(0).view(-1, x.shape[-2])
            attn_res, *_ = layer.attention(h, None, pos_ids)

            # Parallel residual
            h = layer.post_attention_layernorm(x)  # Recomputing from above
            if mlp_eraser is not None:
                h = mlp_eraser(h)

            mlp_res = layer.mlp(h)
            x = x + attn_res + mlp_res
            xs[i] = x.to("cpu", non_blocking=True)

    for batch, x in tqdm(zip(train.iter(batch_size), xs), total=N):
        assert isinstance(batch, dict)
        tokens = assert_type(torch.Tensor, batch["input_ids"])

        x = x.to(model.device)
        x = base.final_layer_norm(x)
        logits = model.embed_out(x)

        labels = tokens.to(logits.device)
        shift_logits = logits[:, :-1, :].contiguous()
        labels = labels[:, 1:].contiguous()
        lm_loss = F.cross_entropy(
            shift_logits.view(-1, shift_logits.size(-1)), labels.view(-1)
        )
        losses.append(lm_loss)

    return scrubber, torch.stack(losses).mean().item()



from contextlib import contextmanager
from functools import partial
from typing import Callable, Literal

import torch
from torch import Tensor, nn
from transformers import PreTrainedModel

from .concept_eraser import ConceptEraser
from .utils import assert_type, is_norm_layer, mangle_module_path


class ConceptScrubber(nn.Module):
    @classmethod
    def from_model(
        cls,
        model: PreTrainedModel,
        z_dim: int = 1,
        affine: bool = True,
        module_suffix: str = "",
        proj_type: Literal["leace", "orth"] = "leace",
        pre_hook: bool = False,
    ):
        d_model = model.config.hidden_size

        scrubber = cls(pre_hook=pre_hook)
        scrubber.erasers.update(
            {
                mangle_module_path(name): ConceptEraser(
                    d_model,
                    z_dim,
                    affine=affine,
                    device=model.device,
                    proj_type=proj_type,
                )
                # Note that we are unwrapping the base model here
                for name, mod in model.base_model.named_modules()
                if is_norm_layer(mod) and name.endswith(module_suffix)
            }
        )
        return scrubber

    def __init__(self, pre_hook: bool = False):
        super().__init__()

        self.erasers = nn.ModuleDict()
        self.pre_hook = pre_hook

    @contextmanager
    def record(
        self,
        model: PreTrainedModel,
        label: Tensor | None = None,
    ):
        """Update erasers with the activations of the model, using the given label.

        This method adds a forward pre-hook to each layer of the model, which
        records its input activations. These are used to update the erasers.
        """

        def record_hook(eraser: ConceptEraser, x: Tensor):
            eraser.update(x, label)

        with self.apply_hook(model, record_hook):
            yield self

    @contextmanager
    def scrub(self, model, dry_run: bool = False):
        """Add hooks to the model which apply the erasers during a forward pass."""

        def scrub_hook(eraser: ConceptEraser, x: Tensor):
            return eraser(x).type_as(x) if not dry_run else x

        with self.apply_hook(model, scrub_hook):
            yield self

    @contextmanager
    def random_scrub(self, model):
        eraser = assert_type(ConceptEraser, next(iter(self.erasers.values())))
        d = eraser.mean_x.shape[0]

        u = eraser.mean_x.new_zeros(d, eraser.z_dim)
        u = nn.init.orthogonal_(u)
        P = torch.eye(d, device=u.device) - u @ u.T

        def scrub_hook(eraser: ConceptEraser, x: Tensor):
            mean = eraser.mean_x

            if eraser.affine:
                _x = (x.type_as(mean) - mean) @ P.T + mean
            else:
                _x = x.type_as(mean) @ P.T

            return _x.type_as(x)

        with self.apply_hook(model, scrub_hook):
            yield self

    @contextmanager
    def apply_hook(
        self,
        model: nn.Module,
        hook_fn: Callable[[ConceptEraser, Tensor], Tensor | None],
    ):
        """Apply a `hook_fn` to each submodule in `model` that we're scrubbing."""

        def post_wrapper(_, __, output, name: str) -> Tensor | None:
            key = mangle_module_path(name)
            eraser = assert_type(ConceptEraser, self.erasers[key])
            return hook_fn(eraser, output)

        def pre_wrapper(_, inputs, name: str) -> tuple[Tensor | None, ...]:
            x, *extras = inputs
            key = mangle_module_path(name)
            eraser = assert_type(ConceptEraser, self.erasers[key])
            return hook_fn(eraser, x), *extras

        # Unwrap the base model if necessary
        if isinstance(model, PreTrainedModel):
            model = model.base_model

        handles = [
            (
                mod.register_forward_pre_hook(partial(pre_wrapper, name=name))
                if self.pre_hook
                else mod.register_forward_hook(partial(post_wrapper, name=name))
            )
            for name, mod in model.named_modules()
            if is_norm_layer(mod) and mangle_module_path(name) in self.erasers
        ]
        assert len(handles) == len(self.erasers), "Not all erasers can be applied"

        try:
            yield self
        finally:
            # Make sure to remove the hooks even if an exception is raised
            for handle in handles:
                handle.remove()



import math

import pytest
import torch
from datasets import Dataset
from transformers import AutoModelForCausalLM, AutoTokenizer

from concept_erasure.scrubbing import patch_attention_, scrub


@pytest.mark.slow
@pytest.mark.parametrize("model_str", ["EleutherAI/pythia-160m", "huggyllama/llama-7b"])
@torch.inference_mode()
def test_scrubbing(model_str: str):
    device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")

    # It turns out to be important to use the actual pretrained weights and not random
    # weights, since random models tend to get near-constant loss even if you implement
    # them incorrectly.
    model = AutoModelForCausalLM.from_pretrained(model_str).to(device)
    tokenizer = AutoTokenizer.from_pretrained(model_str)

    ## Test attention patching by itself ##
    # Get the output pre-patching
    input_ids = tokenizer.encode("Hello world", return_tensors="pt").to(device)
    clean_logits = model(input_ids).logits

    # Patch the model
    patch_attention_(model)

    # Check that the outputs are the same
    patched_logits = model(input_ids).logits
    torch.testing.assert_close(clean_logits, patched_logits, atol=5e-5, rtol=5e-5)

    ## Test attention patching + concept erasure ##
    tokenized = tokenizer(
        ["The quick brown fox jumps over the lazy dog", "Hello world!"]
    )
    ds = Dataset.from_dict(dict(tokenized)).with_format("torch")
    gt_losses = []

    for record in ds:
        input_ids = record["input_ids"].unsqueeze(0).to(device)
        gt_losses.append(model(input_ids, labels=input_ids).loss)

    gt_loss = torch.stack(gt_losses).mean().item()
    _, loss = scrub(model, ds, z_column=None)
    assert math.isclose(gt_loss, loss, rel_tol=1e-4)



from types import MethodType
from typing import Literal

import torch
import torch.nn.functional as F
from datasets import Dataset
from tqdm.auto import tqdm
from transformers import (
    LlamaForCausalLM,
    LlamaModel,
)
from transformers.models.llama.modeling_llama import (
    LlamaAttention,
    LlamaDecoderLayer,
    apply_rotary_pos_emb,
)

from concept_erasure import ConceptEraser, ConceptScrubber
from concept_erasure.utils import assert_type


def _fast_attn(
    self,
    hidden_states: torch.Tensor,
    attention_mask: torch.Tensor | None = None,
    position_ids: torch.LongTensor | None = None,
    past_key_value: tuple[torch.Tensor, ...] | None = None,
    output_attentions: bool = False,
    use_cache: bool = False,
) -> tuple[torch.Tensor, torch.Tensor | None, tuple[torch.Tensor, ...] | None]:
    """Fast version of LLaMA attention."""
    assert not output_attentions, "output_attentions not implemented for fast_attn"
    del attention_mask  # unused, but it's non-None in the base class

    bsz, q_len, _ = hidden_states.size()
    q = (
        self.q_proj(hidden_states)
        .view(bsz, q_len, self.num_heads, self.head_dim)
        .transpose(1, 2)
    )
    k = (
        self.k_proj(hidden_states)
        .view(bsz, q_len, self.num_heads, self.head_dim)
        .transpose(1, 2)
    )
    v = (
        self.v_proj(hidden_states)
        .view(bsz, q_len, self.num_heads, self.head_dim)
        .transpose(1, 2)
    )

    kv_seq_len = k.shape[-2]
    if past_key_value is not None:
        kv_seq_len += past_key_value[0].shape[-2]
    cos, sin = self.rotary_emb(v, seq_len=kv_seq_len)
    q, k = apply_rotary_pos_emb(q, k, cos, sin, position_ids)
    # [bsz, nh, t, hd]

    if past_key_value is not None:
        # reuse k, v, self_attention
        k = torch.cat([past_key_value[0], k], dim=2)
        v = torch.cat([past_key_value[1], v], dim=2)

    past_key_value = (k, v) if use_cache else None
    attn_output = F.scaled_dot_product_attention(q, k, v, is_causal=True)
    attn_output = attn_output.transpose(1, 2)
    attn_output = attn_output.reshape(bsz, q_len, self.hidden_size)
    attn_output = self.o_proj(attn_output)

    return attn_output, None, past_key_value


def patch_attention_llama_(model: torch.nn.Module):
    """Patch the attention layers of a LLaMA model to use fast attention kernels."""
    for mod in model.modules():
        if isinstance(mod, LlamaAttention):
            mod.forward = MethodType(_fast_attn, mod)


@torch.no_grad()
def scrub_llama(
    model: LlamaForCausalLM,
    train: Dataset,
    z_column: str | None,
    batch_size: int = 1,
    proj_type: Literal["leace", "loracs", "orth"] = "leace",
    sublayers: bool = True,
    affine: bool = True,
) -> tuple[ConceptScrubber | None, float]:
    base = assert_type(LlamaModel, model.base_model)
    d = assert_type(int, base.config.hidden_size)

    if z_column is None:
        k = -1
        scrubber = None
    else:
        k = assert_type(int, train.features[z_column].feature.num_classes)
        scrubber = ConceptScrubber()

    losses = []

    xs = []
    zs = []
    N = len(train) // batch_size
    train = train.with_format("torch", device=model.device)

    # Embed all the inputs
    embed_fn = base.get_input_embeddings()
    for i, batch in enumerate(tqdm(train.iter(batch_size), desc="Embedding", total=N)):
        assert isinstance(batch, dict)

        tokens = assert_type(torch.Tensor, batch["input_ids"])
        x = embed_fn(tokens)
        xs.append(x.to("cpu", non_blocking=True))

        # We don't actually need to move these to the CPU since they're small
        if z_column is not None:
            zs.append(F.one_hot(batch[z_column], num_classes=k))

    # Enumerate the layers
    for j, layer in enumerate(tqdm(base.layers, unit="layer")):
        assert isinstance(layer, LlamaDecoderLayer)

        attn_eraser = None
        if scrubber is not None:
            attn_eraser = ConceptEraser(
                d, k, affine=affine, device=model.device, proj_type=proj_type
            )
            scrubber.erasers[f"layers-{j}-input_layernorm"] = attn_eraser

            # Fit the next eraser on the previous hidden states
            for x, z in tqdm(zip(xs, zs), desc="Fitting (attn)", total=N):
                assert scrubber is not None

                # Discard post-LN output and recompute during application to save RAM
                x = layer.input_layernorm(x.to(model.device))
                attn_eraser.update(x, z)

        # Run attention & MLP with the erasers we just fit
        for i, x in tqdm(enumerate(xs), desc="Applying (attn)", total=N):
            # Bring back to the accelerator
            x = x.to(model.device)
            h = layer.input_layernorm(x)  # Recomputing from above

            # Apply the eraser
            if scrubber is not None:
                h = attn_eraser(h).type_as(h)

            pos_ids = torch.arange(0, h.shape[-2], device=h.device, dtype=torch.long)
            pos_ids = pos_ids.unsqueeze(0).view(-1, h.shape[-2])

            h, _, __ = layer.self_attn(h, position_ids=pos_ids)
            h = x = x + h  # Post-attention residual connection

            # We're not scrubbing the sublayers, so run the rest of the layer
            if not sublayers:
                h = layer.post_attention_layernorm(h)
                h = layer.mlp(h)
                h = x + h  # Post-MLP residual connection

            xs[i] = h.to("cpu", non_blocking=True)

        # Skip the part where we scrub the MLP if we're not doing that
        if not sublayers:
            continue

        mlp_eraser = None
        if scrubber is not None:
            mlp_eraser = ConceptEraser(
                d, k, affine=affine, device=model.device, proj_type=proj_type
            )
            scrubber.erasers[f"layers-{j}-post_attention_layernorm"] = mlp_eraser

            # Fit the next eraser on the previous hidden states
            for x, z in tqdm(zip(xs, zs), desc="Fitting (MLP)", total=N):
                assert scrubber is not None

                # Discard post-LN output and recompute during application to save RAM
                x = layer.post_attention_layernorm(x.to(model.device))
                mlp_eraser.update(x, z)

        for i, x in tqdm(enumerate(xs), desc="Applying (MLP)", total=N):
            # Bring back to the accelerator
            x = x.to(model.device)
            h = layer.post_attention_layernorm(x)  # Recomputing from above

            # Apply the eraser
            if scrubber is not None:
                h = mlp_eraser(h).type_as(h)

            h = layer.mlp(h)
            h = x + h  # Post-MLP residual connection
            xs[i] = h.to("cpu", non_blocking=True)

    for batch, x in tqdm(zip(train.iter(batch_size), xs), total=N):
        assert isinstance(batch, dict)
        tokens = assert_type(torch.Tensor, batch["input_ids"])

        x = x.to(model.device)
        x = base.norm(x)
        logits = model.lm_head(x)

        labels = tokens.to(logits.device)
        shift_logits = logits[:, :-1, :].contiguous()
        labels = labels[:, 1:].contiguous()
        lm_loss = F.cross_entropy(
            shift_logits.view(-1, shift_logits.size(-1)), labels.view(-1)
        )
        losses.append(lm_loss)

    return scrubber, torch.stack(losses).mean().item()



from itertools import product

import numpy as np
import pytest
import torch
from sklearn.datasets import make_classification
from sklearn.linear_model import LogisticRegression
from sklearn.svm import LinearSVC

from concept_erasure import ConceptEraser


def test_stats():
    num_features = 3
    num_classes = 2
    batch_size = 10
    num_batches = 5

    eraser = ConceptEraser(num_features, num_classes)

    # Generate random data
    torch.manual_seed(42)
    x_data = [torch.randn(batch_size, num_features) for _ in range(num_batches)]
    y_data = [
        torch.randint(0, num_classes, (batch_size, num_classes))
        for _ in range(num_batches)
    ]

    # Compute cross-covariance matrix using batched updates
    for x, y in zip(x_data, y_data):
        eraser.update(x, y)

    # Compute the expected cross-covariance matrix using the whole dataset
    x_all = torch.cat(x_data)
    y_all = torch.cat(y_data)
    mean_x = x_all.mean(dim=0)
    mean_y = y_all.type_as(x_all).mean(dim=0)
    x_centered = x_all - mean_x
    y_centered = y_all - mean_y

    expected_cov = torch.einsum("b...m,b...n->...mn", x_centered, x_centered)
    expected_cov /= batch_size * num_batches - 1

    expected_sigma_xz = torch.einsum("b...m,b...n->...mn", x_centered, y_centered)
    expected_sigma_xz /= batch_size * num_batches - 1

    # Compare the computed cross-covariance matrix with the expected one
    torch.testing.assert_close(eraser.cov_x, expected_cov)
    torch.testing.assert_close(eraser.sigma_xz, expected_sigma_xz)


# Both `1` and `2` are binary classification problems, but `1` means the labels are
# encoded in a 1D one-hot vector, while `2` means the labels are encoded in an
# n x 2 one-hot matrix.
@pytest.mark.parametrize("num_classes", [1, 2, 3, 5, 10, 20])
@torch.no_grad()
def test_projection(num_classes: int):
    n, d = 2048, 128
    num_distinct = max(num_classes, 2)

    X, Y = make_classification(
        n_samples=n,
        n_features=d,
        n_classes=num_distinct,
        n_informative=num_distinct,
        random_state=42,
    )
    X_t = torch.from_numpy(X)
    Y_t = torch.from_numpy(Y)
    if num_classes > 1:
        Y_1h = torch.nn.functional.one_hot(Y_t, num_classes)
    else:
        Y_1h = Y_t

    eps = 2e-9
    mse_dict: dict[tuple[bool, str], float] = {}

    for affine, proj_type in product([False, True], ["leace", "orth"]):
        eraser = ConceptEraser.fit(X_t, Y_1h, affine=affine, proj_type=proj_type)
        X_ = eraser(X_t)

        # Check idempotence
        torch.testing.assert_close(eraser(X_), X_)

        # Check that the rank of the update <= num_classes
        rank = torch.linalg.svdvals(X_t - X_).gt(eps).sum()
        assert rank <= num_classes

        # Record the mean squared error for comparison
        X_np = X_.numpy()
        mse_dict[(affine, proj_type)] = np.square(X_np - X).mean()

        # Check that the unconditional mean is unchanged
        if affine:
            torch.testing.assert_close(X_t.mean(dim=0), X_.mean(dim=0))

        # Compute class means and check that they are equal after the projection
        class_means_ = torch.stack(
            [X_[Y_t == c].mean(dim=0) for c in range(num_distinct)]
        )
        torch.testing.assert_close(class_means_[1:], class_means_[:-1])

        # Sanity check that class means are NOT equal before the projection
        class_means = torch.stack(
            [X_t[Y_t == c].mean(dim=0) for c in range(num_distinct)]
        )
        assert not torch.allclose(class_means[1:], class_means[:-1])

        # Logistic regression should not be able to learn anything.
        null_lr = LogisticRegression(penalty=None, tol=0.0).fit(
            # Weirdly, in order for this to work consistently with solver='lbfgs', we
            # need to center the design matrix first; otherwise the coefficients aren't
            # quite small enough. Other solvers don't have this problem but they're
            # slow. In theory, the centroid of X shouldn't matter to the solution.
            X_.numpy() - X_.numpy().mean(axis=0),
            Y,
        )
        assert abs(null_lr.coef_).max() < eps

        # Sanity check that it DOES learn something before the projection
        real_lr = LogisticRegression(penalty=None, tol=0.0).fit(
            # Do the same centering operation here to be consistent
            X - X.mean(axis=0),
            Y,
        )
        assert abs(real_lr.coef_).max() > 0.1

        # Linear SVM shouldn't be able to learn anything either
        null_svm = LinearSVC(
            # The dual formulation injects random noise into the solution
            dual=False,
            # Unfortunately the intercept is subject to L2 regularization; setting this
            # to a large value largely cancels out the effect. Regularizing the
            # intercept can cause the coefficients to get larger than they should be.
            intercept_scaling=1e6,
            tol=eps,
        ).fit(X_np, Y)
        assert abs(null_svm.coef_).max() < eps

        # But it should learn something before the projection
        real_svm = LinearSVC(dual=False, intercept_scaling=1e6, tol=eps).fit(X, Y)
        assert abs(real_svm.coef_).max() > 0.1

    # Check that using proj_type="leace" strictly better than "orth"
    assert mse_dict[(True, "leace")] < mse_dict[(True, "orth")]
    assert mse_dict[(False, "leace")] < mse_dict[(False, "orth")]



from .concept_eraser import ConceptEraser
from .concept_scrubber import ConceptScrubber
from .data import chunk_and_tokenize
from .utils import assert_type, chunk

__all__ = [
    "assert_type",
    "chunk",
    "chunk_and_tokenize",
    "ConceptEraser",
    "ConceptScrubber",
]



import math
from typing import Any, Type, TypeVar, cast

from torch import nn
from transformers import PreTrainedModel

T = TypeVar("T")


def assert_type(typ: Type[T], obj: Any) -> T:
    """Assert that an object is of a given type at runtime and return it."""
    if not isinstance(obj, typ):
        raise TypeError(f"Expected {typ.__name__}, got {type(obj).__name__}")

    return cast(typ, obj)


def chunk(seq: list[T], chunk_size: int) -> list[list[T]]:
    """Chunk a sequence into chunks of size `chunk_size`."""

    # Why the hell is this not in the standard library?
    return [
        seq[i * chunk_size : (i + 1) * chunk_size]
        for i in range(math.ceil(len(seq) / chunk_size))
    ]


def get_transformer_layers(model: PreTrainedModel) -> nn.ModuleList:
    """Return the `nn.ModuleList` containing the transformer layers in a model."""
    assert not model.config.is_encoder_decoder, "Encoder-decoder models not supported"

    lists = [mod for mod in model.modules() if isinstance(mod, nn.ModuleList)]
    if not lists:
        raise ValueError("Could not find transformer layers")

    # Return the module list with the most parameters
    return max(lists, key=lambda mod: sum(p.numel() for p in mod.parameters()))


def is_norm_layer(module: nn.Module) -> bool:
    """Return `True` if the module is a normalization layer."""
    cls_name = type(module).__name__
    return cls_name.endswith("LayerNorm") or cls_name.endswith("RMSNorm")


def mangle_module_path(name: str) -> str:
    """Mangle a module path to make it a valid key in a `nn.ModuleDict`."""
    # This is a weird edge case we probably don't need to support
    assert "-" not in name, "Module path cannot contain `-`"
    return name.replace(".", "-")
